% SVM with Stoachstic Gradient Descent. Parallel implementation for Matlab.
% MeanSgd according to Rohrbach et al., CVPR 2011, 
% Evaluating Knowledge Transfer and Zero-Shot Learning in a 
% Large-Scale Setting
%
% Copyright (C) 2011 Marcus Rohrbach
% Implementation of SGD based on code by Leon Bottou
% see http://leon.bottou.org/projects/sgd.
% 
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.

classdef meanSgd
    %SGD Stochastic gradient descent according to Lon Bottou
  
  properties
    libPath = 'bin/';
    functionName = 0;
    options;
  end
  
  methods
    function obj = meanSgd(varargin)
    end
    
    function fName = getFunctionName(obj, classId)
      if ~obj.functionName(1)
        obj.functionName = [obj.options.svm_types{obj.options.svm_type} obj.options.loss];
      end
      if nargin < 2 || strcmp(classId, 'double')
        fName = obj.functionName;
      else
        fName = [obj.functionName 'Single'];
      end
      
      if obj.options.multiclass
        fName = [fName 'Openmp'];
      end
    end
    
    function str = getTestParamsIdString(obj)
      str = ['epoch' num2str(obj.options.testEpoch)];
      if obj.options.meanEpochs
        str = [str 'mean' obj.options.meanEpochFactor];
      end
      
      str = [str getTestParamsIdString@SvmTrainer(obj)];
    end
    
    function str = getModelParamsIdString(obj)
      str = [obj.getFunctionName() 'Epochs' num2str(obj.options.epochs)];
      str = [str 'Lambda' num2str(obj.options.lambda)];
      
      if ~strcmp(obj.options.costEstimation, 'none')
        str = [str obj.options.costEstimation];
      end
      
      if obj.options.randomize > 0
        str = [str 'Rand' int2str(obj.options.randomize)];
      end
      if obj.options.continueTemp
        str = [str 'ContT'];
      end
      str = [str getModelParamsIdString@SvmTrainer(obj)];
    end
    
    function fdp = addTrainerName(obj, fdp)
      [pathstr, n, ext] = fileparts(fdp);
      if obj.options.kernelType > 0;
        fdp = [pathstr filesep n '.' obj.name ...
          '-'  obj.getKernelName() '-' getKernelParamsString(obj) ext];
      else
        fdp = [pathstr filesep n '.' obj.name ext];
      end
    end
    
    function [model svmModels] = runTrain(obj, featureMatrix, labelVector)
      nClasses = size(labelVector,2);
      
      labelVector=double(labelVector);
      
      % check inputs
      if min(labelVector) < 0
        labelVector = labelVector>0;
      end
      if issparse(featureMatrix)
        error('feature matrix must not be sparse');
      end
      floatType =class(featureMatrix);
      fprintf('Using function %s\n', obj.getFunctionName(floatType));
      
      svmModels = cell(nClasses,1);
      for iEpoch = 1:obj.options.epochs
        if iEpoch == 1
          [model.w model.bias temperature] = ...
            feval(obj.getFunctionName(floatType), featureMatrix, labelVector, ...
            obj.options.lambda, 1, obj.options.randomize);
        else
          if obj.options.continueTemp
            [model.w model.bias temperature] = ...
              feval(obj.getFunctionName(floatType), featureMatrix, labelVector, ...
              obj.options.lambda, 1, obj.options.randomize, model.w, model.bias, temperature);
          else
            [model.w model.bias] = ...
              feval(obj.getFunctionName(floatType), featureMatrix, labelVector, ...
              obj.options.lambda, 1, obj.options.randomize, model.w, model.bias);
          end
        end
        if obj.options.saveAllModelsIndividually
          for iClass = 1:nClasses
            svmModels{iClass}.w = model.w(:,iClass);
            svmModels{iClass}.bias = model.bias(iClass);
            svmModels{iClass}.wCell{iEpoch} = svmModels{iClass}.w;
            svmModels{iClass}.biasCell{iEpoch} = svmModels{iClass}.bias;
          end
        end
        model.wCell{iEpoch} = model.w;
        model.biasCell{iEpoch} = model.bias;
      end
      
      
    end
    
    %% Prediction
    function confidence = predict(obj, featureMatrix, svmModel) 
      
      if obj.options.testEpoch > 0
        if obj.options.meanEpochs
          wSum =  0;
          biasSum = 0;
          
          sumFactor = 0;
          for iEpoch = 1:obj.options.testEpoch
            switch(obj.options.meanEpochFactor)
              case 'linear'
                factor = iEpoch;
              case ''
                factor = 1;
              otherwise
                error('unkown svm meanEpochFactor')
            end
            wSum = wSum + factor * svmModel.wCell{iEpoch};
            biasSum = biasSum + factor * svmModel.biasCell{iEpoch};
            sumFactor = sumFactor + factor;
          end
          
          wSum = wSum / sumFactor;
          biasSum = biasSum / sumFactor;
          
          confidence = bsxfun(@plus, (wSum' * featureMatrix)', biasSum);
          
        else
          confidence = bsxfun(@plus, (svmModel.wCell{obj.options.testEpoch}' * featureMatrix)',...
            svmModel.biasCell{obj.options.testEpoch});
        end
        
        
      else
        if isfield(svmModel, 'wscale')
          confidence = bsxfun(@plus, (bsxfun(@times,...
            svmModel.w, svmModel.wscale)' * featureMatrix)', svmModel.bias);
        else
          confidence = bsxfun(@plus, (svmModel.w' * featureMatrix)', svmModel.bias);
        end
      end
    end
    
    %% options
    function options = getDefaultOptions(obj)
      options.svm_type = 1;
      options.svm_types = {'meanSvmsgd2'};
      
      options.epochs = 20;
      options.testEpoch = 20;
      options.meanEpochs = 1;       % 1 for MeanSgd, 0 for Sgd
      options.lambda = 1e-7;
      options.randomize = 1;        % 1 for *Stochastic* gradient descent
      options.continueTemp = false;  % continue to decrease temperature after each epoch
      options.saveAllModelsIndividually = true;
      
      % prediction parameteres
      options.meanEpochFactor = ''; % '': all epochs are weighted equally
      
      options.loss = 'Hingeloss';
      
      options.multiclass = true; % Enable parrallel compuation
    end
  end
  %%
  methods (Static)
    function test()
      % small test with two different and overlapping label sets.
      % this occures, e.g. when training for different, overlapping classes
      % such as attributes.
      % works also for multi class setting.
      % note that parameters are not specifically tuned for this setting
      % an meanSgd is especially succesfull for large datasets.
      % We used it for about 1.2 million image samples.
      
      sgdTrainer = meanSgd();
      
      sgdTrainer.options = sgdTrainer.getDefaultOptions();
      sgdTrainer.options.lambda = 1e-4;
      addpath(sgdTrainer.libPath)
      
      
      
      [featureMatrix labelVectors] = meanSgd.createAndPlotData();

      % training
      [model models] = sgdTrainer.runTrain(featureMatrix, labelVectors);
      
      % show models
      meanSgd.plotModel(models, featureMatrix);
      
      % test (here on training data)
      predictedConfidences = sgdTrainer.predict(featureMatrix, models{1});
    end
    
    
    function [featureMatrix labelVectors] = createAndPlotData()
      figure;
      hold on;
      nTraining = 1000;

      meanFactor = 3;
      sigma = 0.75;
      
      trainingData= cell(2,2);
      trainingLabels = cell(2,2);
      for x = 0:1
        for y = 0:1
          trainingData{x+1,y+1} = [sigma*randn(nTraining,1)+meanFactor*x, ...
            sigma*randn(nTraining,1)+meanFactor*y+1];
          
          trainingLabels{x+1,y+1} = [zeros(nTraining, 1) + x, zeros(nTraining, 1) + y];
          
          if y
            label = 'r';
          else
            label = 'g';
          end
          if x
            label = [label 'x']; %#ok<AGROW>
          else
            label = [label 'o']; %#ok<AGROW>
          end
          plot(trainingData{x+1,y+1}(:,1),trainingData{x+1,y+1}(:,2) , label);
        end
      end
      
      
      hold off;
      
      featureMatrix = single(cell2mat(trainingData(:))');
      labelVectors = cell2mat(trainingLabels(:));
    end
    
    function plotModel(models, featureMatrix)
            title(['Training data and model. MeanSgd: solid, Sgd: dashed'...
        '\newline black lines separate symbols, blue lines are learned on color labels'...
        '\newline \fontsize{8} \it Note that parameters are not tuned for this toy example']);
      hold on;
      x = -3:0.01:6;
      iModel = 1;
      
      
      y = x*models{iModel}.w(1)/models{iModel}.w(2)+models{iModel}.bias/models{iModel}.w(2);
      
      meanW = mean(cell2mat(models{iModel}.wCell)');
      meanBias = mean(cell2mat(models{iModel}.biasCell));
      meanY = x*meanW(1)/meanW(2)+meanBias/meanW(2);
      
      plot(x,y,'k--');
      plot(x,meanY,'k-');
      iModel = 2;
      
      
      y = x*models{iModel}.w(2)/models{iModel}.w(1)+models{iModel}.bias/models{iModel}.w(1);
      
      meanW = mean(cell2mat(models{iModel}.wCell)');
      meanBias = mean(cell2mat(models{iModel}.biasCell));
      meanY = x*meanW(2)/meanW(1)+meanBias/meanW(1);
      
      plot(y,x,'b--');
      plot(meanY,x,'b-');
      
      xlimits =[min(featureMatrix(1,:)) max(featureMatrix(1,:))];
      ylimits =[min(featureMatrix(2,:)) max(featureMatrix(2,:))];
      xlim(xlimits);
      ylim(ylimits);
      hold off;
    end
  end
end
