//    SVM with Stoachstic Gradient Descent. Parallel implementation for Matlab.
//    Copyright (C) 2011 Marcus Rohrbach
//    Implementation of SGD based on code by Leon Bottou
//    see http://leon.bottou.org/projects/sgd.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.


#include "math.h"
#include "mex.h"
#include <omp.h>

// floating point type of data matrix
#define floatType float

#define HINGELOSS 1
#define SMOOTHHINGELOSS 2
#define SQUAREDHINGELOSS 3
#define LOGLOSS 10
#define LOGLOSSMARGIN 11

// Select loss
#define LOSS HINGELOSS

// Zero when no bias, one when bias term should be used
#define BIAS 1

#include "helpers.h"

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	// check for correct input
	if (nrhs < 5 || nlhs < 2) {
		mexPrintf("Error: %d inputs and %d outputs provided\n", nrhs, nlhs);
		mexPrintf("5 or 8 inputs and 2-3 outputs required:");
		mexPrintf("[w, bias, wscale] = svmsgd(featureMatrix, labelVector, lambda, epochs, randomize)");
		mexPrintf("[w, bias, wscale] = svmsgd(featureMatrix, labelVector, lambda, epochs, randomize, w0, bias0, t0)");
		mexPrintf("");
		mexPrintf("Outputs:");
		mexPrintf("w:             double [feat dimension, #classes]");
		mexPrintf("bias:          double [1, #classes]");
		mexPrintf("t:             double [1, #classes], optional");
		mexPrintf("");
		mexPrintf("\nInputs:");
		mexPrintf("featureMatrix: floatType [feat dimension, #data points]");
		mexPrintf("labelVector:   double [#classes, #data points]");
		mexPrintf("lambda:        double");
		mexPrintf("epochs:        int");
		mexPrintf("randomize:     >0: randomly order data points/lables");
		mexPrintf("w0:            double [feat dimension, #classes], optional");
		mexPrintf("bias0:         double [1, #classes], optional");
		mexPrintf("t0:            double [1, #classes], optional");
		return;
	}

	// parse inputs
	const mxArray* xDataArray = prhs[0];
	const mxArray* yDataArray = prhs[1];
	const floatType* xData = (floatType*)mxGetData(xDataArray);
	const double* yData = mxGetPr(yDataArray);

	const double lambda = mxGetScalar(prhs[2]);
	const mwSize epochs = (mwSize) mxGetScalar(prhs[3]);
	const double randomize = mxGetScalar(prhs[4]);

	const mwSize featDim = mxGetM(xDataArray); // # rows
	const mwSize nData = mxGetN(xDataArray); // # cols
	const mwSize nLabels = mxGetM(yDataArray); // # rows
	const mwSize nClasses = mxGetN(yDataArray); // # cols

	// display statistics of input.
	mexPrintf("featDim:     %lu\n", featDim);
	mexPrintf("nData:       %lu\n", nData);
	mexPrintf("nClasses:    %lu\n", nClasses);
	mexPrintf("nLabels:     %lu\n", nLabels);
	mexPrintf("Max threads: %d\n", omp_get_max_threads());
	mexPrintf("randomize:   %f\n", randomize);
	mexPrintf("type of data:%s\n", mxGetClassName(xDataArray));
	mexPrintf("lambda:      %g\n", lambda);

	// check input dimensions
	if (nLabels != nData) {
		mexErrMsgTxt("Second Dimension of featureMatrix and labelVector do not correspond");
	}

	// weight w
	mxArray* wArray;
	if (nrhs > 5) {
		wArray = mxDuplicateArray(prhs[5]);
		mexPrintf("using provided initialization values for weight vectors.\n");
	} else {
		wArray = mxCreateDoubleMatrix(featDim, nClasses, mxREAL);
		mexPrintf("initializing weights with zero.\n");
	}
	double* w = mxGetPr(wArray);

	// bias
	mxArray* biasArray;
	if (nrhs > 6) {
		biasArray = mxDuplicateArray(prhs[6]);
	} else {
		biasArray = mxCreateDoubleMatrix(1, nClasses, mxREAL);
	}
	double* bias = mxGetPr(biasArray);



	// scaling factor for w
	mxArray* wscaleArray = mxCreateDoubleMatrix(1, nClasses, mxREAL);
	double* wscale = mxGetPr(wscaleArray);
	for (mwSize iClass = 0; iClass < nClasses; iClass++) {
		wscale[iClass]=1;
	}

	// init constants
	const double maxw = 1.0 / sqrt(lambda);
	const double typw = sqrt(maxw);
	const double eta0 = typw / max(1.0,  dloss(-typw));

	// temperature
	mxArray* tArray;
	double* t;
	if (nrhs > 7) {
		// continue with old temperature
		tArray = mxDuplicateArray(prhs[7]);
		t = mxGetPr(tArray);
	} else {
		// init temperature
		tArray = mxCreateDoubleMatrix(1, nClasses, mxREAL);
		t = mxGetPr(tArray);
		double t0 = 1.0 / (eta0 * lambda);
		for (int iClassInt=0; iClassInt < (int) nClasses; iClassInt++) {
			t[iClassInt] = t0;
		}
	}


	// create random ordering
	mxArray* iDataRandPermArray[1];
	mxArray* nDataArray[1];
	double* iDataRandPerm;
	if (randomize > 0) {
		nDataArray[0] = mxCreateDoubleScalar((double) nData);
		mexCallMATLAB(1,iDataRandPermArray,1,nDataArray,"randperm");
		iDataRandPerm = mxGetPr(iDataRandPermArray[0]);
	}

	mexPrintf("initial t    %g\n", t[0]);
	mexPrintf("initial eta0 %g\n", eta0);

	mexPrintf("finished setup, starting parallel computation\n");

	// parallel SGD for all classes
#pragma omp parallel for schedule(dynamic)
	for (int iClassInt=0; iClassInt < (int) nClasses; iClassInt++)  {
		mwSize iClass = iClassInt;

		mwSize iData;
		for (mwSize iEpoch = 0; iEpoch < epochs; iEpoch++) {

			for (mwSize i = 0; i < nData; i++) {

				if (randomize > 0) {
					// subtract one from matlab to c indexes
					iData = (mwSize) iDataRandPerm[i] -1;
				} else {
					iData = i;
				}

				double y = yData[iClass*nData+iData];
				// ignore -2 and smaller values
				if (y >= -1) {
					y = (y>0) ? 1.0 : -1.0;


					double eta = 1.0 / (lambda * t[iClass]);
					double s = 1 - eta * lambda;
					wscale[iClass] *= s;
					if (wscale[iClass] < 1e-9)
					{
						for (mwSize iFeat = 0; iFeat < featDim; iFeat++) {
							w[iClass*featDim+iFeat] *= wscale[iClass];
						}
						wscale[iClass] = 1;
					}

					double wx = inner_product(&w[iClass*featDim], &xData[iData*featDim], featDim) * wscale[iClass];
					double z = y * (wx + bias[iClass]);
#if LOSS < LOGLOSS
					if (z < 1)
#endif
					{
						double etd = eta * dloss(z);
						// w += xData[:,i] * etd * y
						for (mwSize iFeat = 0; iFeat < featDim; iFeat++) {
							w[iClass*featDim+iFeat] += static_cast<double>(xData[iData*featDim +iFeat]) * etd * y / wscale[iClass];
						}
#if BIAS
						bias[iClass] += etd * y * 0.01;
#endif
					}
					t[iClass] += 1;
				}
			}

			// scale w
			for (mwSize iFeat = 0; iFeat < featDim; iFeat++) {
				w[iClass*featDim+iFeat] *= wscale[iClass];
			}
			wscale[iClass] = 1;
		}
	}

	mexPrintf("finished parallel computation!\n");

	// set return values
	plhs[0] = wArray;
	plhs[1] = biasArray;
	plhs[2] = tArray;
	return;
}
